
local addon_name, ns = ...
local rotation = ns.rotation
local rearrange = ns.rearrange
local FramePool = ns.FramePool
local settings = ns.settings
local lib = ns.lib

local Printer = LibStub:GetLibrary("OakPrint-1.0")(addon_name)

local display = {}
local opts = settings.display or {}
--------------------------------------------------------------------------------
local GetTime = GetTime

local LDB = LibStub("LibDataBroker-1.1")
local dataobj = LDB:GetDataObjectByName("OakInterrupts")

local function indexOf(t, val)
    for i, v in ipairs(t) do
        if v == val then
            return i
        end
    end
end

--------------------------------------------------------------------------------
-- Lookup tables
--------------------------------------------------------------------------------

local class = setmetatable({}, { __index = function(t, k)
        local cls = select(2, UnitClass(k))
        t[k] = cls
        return cls
    end})

local displayof = {}
local DispByName = {}

--------------------------------------------------------------------------------
-- Interrupter display functions
--------------------------------------------------------------------------------

local default_colour = ns.settings.colour.default
local highlight_colour = ns.settings.colour.highlight

local function SetNameText(frame, name, class)
    local colour = "|c" .. lib.classHexColour(class)
    frame.txtname:SetText(colour .. lib.formatName(name) .. "|r")
end

local function SetTimeText(frame, time)
    local colour
    local t = (time == "Polling" and "Polling") or lib.formatTime(time)
    if t == "READY" then
        colour = "|cFF00FF00"
    elseif t == "Polling" then
        colour = "|cFF33AACC"
    else
        colour = "|cFFFF0000"
    end
    frame.time:SetText(colour .. t .. "|r")
end

local function SetInterrupter(frame, name)
    frame.name = name
    local class = class[name]
    SetNameText(frame, name, class)
end

local function OnMouseDown(frame, btn)
    if btn == "RightButton" then
        local disp = displayof[frame]
        disp:RemoveInterrupter(frame)
        disp.queue:LayoutQueue()
    elseif btn == "LeftButton" then
        rearrange.BeginMoving(displayof[frame].queue, frame, btn)
    end
end

local function OnMouseUp(frame, btn)
    if btn == "LeftButton" then
        rearrange.EndMoving(displayof[frame].queue, frame, btn)
    end
end

local function OnUpdate(frame, elapsed)
    rearrange.MouseUpdate(displayof[frame].queue, frame, btn)
    -- update time display
    if dataobj[frame.name] then
        SetTimeText(frame, dataobj[frame.name] - GetTime())
    end
end

--------------------------------------------------------------------------------
-- FramePools
--------------------------------------------------------------------------------

local display_pool = FramePool:NewPool("OakInterrupts_Display", UIParent)
local interrupter_pool = FramePool:NewPool("OakInterrupts_InterrupterFrame", UIParent)
--------------------------------------------------------------------------------
-- Create interrupter frames
--------------------------------------------------------------------------------

local function LayoutFrame(disp, frame)
    frame:SetWidth(disp.width - 2 * disp.padding)
    frame:SetHeight(disp.player_height)
end

local function InitFrame(disp, frame, name)
    local bg = frame:CreateTexture()
    bg:SetAllPoints()
    bg:SetTexture(unpack(default_colour))
    frame.bg = bg

    local txtname = frame:CreateFontString()
    txtname:SetPoint("LEFT", frame, "LEFT", 5, 0)
    txtname:SetFontObject("GameFontNormal")
    frame.txtname = txtname

    local time = frame:CreateFontString()
    time:SetPoint("RIGHT", frame, "RIGHT", -5, 0)
    time:SetFontObject("GameFontNormal")
    frame.time = time

    frame:EnableMouse(true)
    frame:SetMovable(true)
    frame:SetScript("OnUpdate", OnUpdate)
    frame:SetScript("OnMouseDown", OnMouseDown)
    frame:SetScript("OnMouseUp", OnMouseUp)
end

local function CreateInterrupterFrame(disp, name)
    -- call disp.queue.LayoutQueue after this returns if you want to update
    -- the display.
    local frame, new = interrupter_pool:NewFrame()--disp.pool:NewFrame()
    if new then
        InitFrame(disp, frame, name)
    end
    frame:SetParent(disp.container)
    LayoutFrame(disp, frame)
    frame:Show()
    SetInterrupter(frame, name)
    return frame
end


--------------------------------------------------------------------------------
-- Base frames for a display
--------------------------------------------------------------------------------

-- methods --
local function AddInterrupter(disp, name)
    if name == "type" then return end
    assert(not disp.watch_list[name], string.format("Error: %s is already watching %s", disp.name, name))
    Printer.debug(1, "%s is now watching %s", disp.name, name)
    local frame = CreateInterrupterFrame(disp, name)
    disp.watch_list[name] = frame
    disp.queue:AddFrame(frame)
    displayof[frame] = disp
    frame:Show()
end

local function RemoveInterrupter(disp, nameorframe)
    assert(type(nameorframe) == "string" or type(nameorframe) == "table",
                "Usage: RemoveInterrupter(display, name | frame)")
    local frame
    local name
    if type(nameorframe) == "string" then
        for _, f in ipairs(disp.queue.frames) do
            if f.name == nameorframe then
                frame = f
                name = nameorframe
            end
        end
    elseif type(nameorframe) == "table" then
        frame = nameorframe
        name = frame.name
    end
    Printer.debug(3, "removing interrupter - frame: %s, name: %s", frame:GetName() or tostring(frame), name)

    disp.watch_list[name] = false
    disp.queue:RemoveFrame(frame)
    displayof[frame] = nil
    Printer.debug(1, "%s no longer watching %s", disp.name, name)
    interrupter_pool:Recycle(frame)
end

local function DataChanged(disp, eventname, doname, attr, value, dobj)
    if not dataobj[attr] then
        disp:RemoveInterrupter(attr)
        disp.watch_list[attr] = nil
        disp.queue:LayoutQueue()
    elseif disp.watch_list[attr] == false then
        return
    elseif disp.watch_list[attr] and value > GetTime() then
        Printer.debug(3, "%s used interrupt", attr)
        rotation.AdvanceRotation(disp, attr, dataobj)
    else -- we are not explicitly ignoring or watching
        if disp.autoinclude then
            disp:AddInterrupter(attr)
            disp.queue:LayoutQueue()
        end
    end
end

local function ChangeName(disp, name)
    if DispByName[name] then
        Printer.print("You already have an interrupt list with the name %s", name)
        return
    end
    disp.heading.text:SetText(name)
    DispByName[disp.name] = nil
    disp.name = name
    DispByName[name] = disp
end

local function Refill(disp)
    for k, v in LDB:pairs(dataobj) do
        if not disp.watch_list[k] then
            disp:AddInterrupter(k)
        end
    end
    disp.queue:LayoutQueue()
end

-- layout methods --
local function LayoutHeading(disp)
    local heading, anchor, pad = disp.heading, disp.queue.anchor, disp.padding
    if anchor == "LEFT" or anchor == "RIGHT" then
        heading:SetWidth(#disp.queue.frames * (disp.width - pad) + pad)
    else
        heading:SetWidth(disp.width)
    end
end

-- creating a display --
local function CreateDisplay(name, height, width)
    local padding = opts.padding
    local height = height or opts.player_height
    local width = width or opts.player_width

    local display, new = display_pool:NewFrame()
    do
        local bg = display:CreateTexture()
        bg:SetAllPoints()
        bg:SetTexture(.1,.1,.1,.7)
    end

    local heading, container

    if new then
        -- create a new one
        heading = CreateFrame("Frame", nil, display)
        container = CreateFrame("Frame", nil, display)

        local text = heading:CreateFontString()
        text:SetPoint("CENTER")
        text:SetFontObject("GameFontNormal")
        heading.text = text
    else
        heading = display.heading
        container = display.container
    end

    -- Initialise display --
    ------------------------

    -- heading
    heading:SetWidth(width)
    heading:EnableMouse(true)
    heading:CreateTitleRegion():SetAllPoints()

    heading.text:SetText(name or "Interrupters")

    -- layout components
    do
        local hh = heading.text:GetStringHeight()
        heading:SetHeight(hh)
        display:SetHeight(hh + 2 * padding)

        heading:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 50, -50)
        display:SetPoint("TOP", heading, "TOP", 0, padding)
        display:SetPoint("BOTTOM", container, "BOTTOM", 0, -padding)
        display:SetPoint("LEFT", heading, "LEFT", 0, 0)
        display:SetPoint("RIGHT", heading, "RIGHT", 0, 0)
        container:SetPoint("TOP", heading, "BOTTOM")
    end

    -- layout options
    display.heading = heading
    display.queue = ns.FrameQueue(container, true)

    display.padding = padding
    display.player_height = height
    display.width = width
    display.name = name
    display.autoinclude = false
    --display.pool = FramePool:NewPool(nil, container)
    display.container = container

    display.watch_list = {} -- the subset of dataobj this display watches

    -- methods
    display.AddInterrupter = AddInterrupter
    display.RemoveInterrupter = RemoveInterrupter
    display.LayoutHeading = LayoutHeading
    display.DataChanged = DataChanged
    display.ChangeName = ChangeName
    display.Refill = Refill

    return display
end

--------------------------------------------------------------------------------
-- Interrupter control
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Display interface
--------------------------------------------------------------------------------

local function ProcessDisp(disporname)
    if type(disporname) == "string" then
        return DispByName[disporname]
    else
        return disporname
    end
end

function display.Lock(disp, mode)
    disp = ProcessDisp(disp)
    if not disp then return end
    Printer.debug(2, "Locking display %s", disp.name)
    disp.heading:EnableMouse(false)
    if mode and string.upper(mode) == "COMBAT" then
        for _, f in ipairs(disp.queue.frames) do
            f:EnableMouse(false)
        end
    else
        disp.locked = true
    end
end

function display.Unlock(disp, mode)
    disp = ProcessDisp(disp)
    if not disp then return end
    Printer.debug(2, "Unlocking display %s", disp.name)
    if not display.locked then
        disp.heading:EnableMouse(true)
    end
    if mode and string.upper(mode) == "COMBAT" then
        for _, f in ipairs(disp.queue.frames) do
            f:EnableMouse(true)
        end
    end
end

function display.SetDirection(disp, dir)
    disp = ProcessDisp(disp)
    if not disp then return end
    anchor = string.upper(dir)
    -- change this when moving to FrameQueue2
    disp.queue.anchor = anchor
    disp:SetPoint("TOP", disp.container, "TOP")
    disp:SetPoint("BOTTOM", disp.heading, "BOTTOM")
    LayoutHeading(disp)
    disp.queue:LayoutQueue()
end

function display.SetAutoInclude(disp, b)
    assert(type(b) == "string" or type(b) == "boolean",
        "Usage: SetAutoInclude(disp, boolean), to boolean can be given as a string.")
    disp = ProcessDisp(disp)
    if type(b) == "string" then
        if string.upper(b) == "FALSE" then
           b = false
       else
           b = true
       end
    end
    disp.autoinclude = b
end

function display.Rename(disp, name)
    assert(type(name) == "string", "Usage: Rename(display, name) - must provide a string name")
    disp = ProcessDisp(disp)
    disp:ChangeName(name)
end

local count = 0
function display.NewDisplay(name)
    if DispByName[name] then
        Printer.print("You already have an interrupt list with the name %s", name)
        return
    end
    count = (not name and count + 1) or count
    local disp = CreateDisplay(name or "Interrupters" .. count)
    for k, _ in LDB:pairs(dataobj) do
        disp:AddInterrupter(k)
    end
    disp:Show()
    disp.queue:LayoutQueue()
    LDB.RegisterCallback(disp, "LibDataBroker_AttributeChanged_OakInterrupts", "DataChanged")
    DispByName[name] = disp
end

function display.CloseDisplay(disp)
    disp = ProcessDisp(disp)
    if not disp then return end
    Printer.debug(1, "Closing display %s", disp.name)
    LDB.UnregisterCallback(disp, "LibDataBroker_AttributeChanged_OakInterrupts")
    for k, v in ipairs(disp.queue.frames) do
        RemoveInterrupter(disp, v)
    end
    DispByName[disp.name] = nil
    display_pool:Recycle(disp)
end

function display.Refill(disp)
    disp = ProcessDisp(disp)
    Printer.debug(1, "Refilling display %s", disp.name)
    disp:Refill()
end

function display.EnableRotation(disp, flag)
    disp = ProcessDisp(disp)
    local announce = flag == "ANNOUNCE" and true or false
    Printer.debug(2, "Enabling rotation for %s: %s", disp.name, tostring(announce))
    if flag then
        disp.rotation = { next_index = 1, announce = announce }
    else
        disp.rotation = nil
    end
end

function display.ShareRotation(disp)
    disp = ProcessDisp(disp)
    if disp.rotation then
        rotation.AnnounceOrder(disp)
    end
end

--------------------------------------------------------------------------------
ns.displays = DispByName
ns.display = display
ns.Printer = Printer

-- TODO: Announce/rotation advancement stuff
--       Drag-to-rearrange functionality moved here from FrameQueue (it is currently all commented out there)
--       deal with having the same person in multiple lists (maybe don't allow it)
